#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include "serial.h"

#undef serial_config_create
#undef serial_has_data

struct serial {
	int fd;
	serial_handler_t handler;
};


serial_t *serial_init()
{
	serial_t *s;

	if (!(s = (serial_t *)malloc(sizeof(struct serial)))) return 0;
	memset(s, 0, sizeof(struct serial));
	s->fd = -1;

	return s;
}


serial_handler_t serial_notification_handler(serial_t *serial, serial_handler_t handler)
{
	serial_handler_t old;

	old = serial->handler;
	serial->handler = handler;

	return old;
}


struct serial_speed {
	int id;
	int baudrate;
};

struct serial_speed serial_speeds[] = {
	{ B0, 0 },
	{ B50, 50 },
	{ B75, 75 },
	{ B110, 110 },
	{ B134, 134 },
	{ B150, 150 },
	{ B200, 200 },
	{ B300, 300 },
	{ B600, 600 },
	{ B1200, 1200 },
	{ B1800, 1800 },
	{ B2400, 2400 },
	{ B4800, 4800 },
	{ B9600, 9600 },
	{ B19200, 19200 },
	{ B38400, 38400 },
	{ B57600, 57600 },
	{ B115200, 115200 },
	{ B230400, 230400 },

	{ -1, -1 }
};


int internal_serial_speed(int baudrate)
{
	struct serial_speed *p;

	for (p = serial_speeds; p->baudrate != -1; p++) {
//		printf("id=%d; baudrate=%d;\n", p->id, p->baudrate);
		if (p->baudrate == baudrate) return p->id;
	}

	return -1;
}


int serial_open(serial_t *serial, const char *device, serial_config_t *config)
{
	struct termios tconfig;
    	int flags;
    	int speed;

    	if ((speed = internal_serial_speed(config->speed)) < 0) {
    		return -1;
    	}

    	switch (config->word_size) {
    		case 8:
    			flags = CS8;
    			break;
    		case 7:
    			flags = CS7;
    			break;
    		case 6:
    			flags = CS6;
    			break;
    	}

    	switch (config->stop_bits) {
    		case 1:
    			break;
    		case 2:
    			flags |= CSTOPB;
    			break;
    	}

    	switch (config->parity) {
		case SERIAL_PARITY_NONE:
			break;

		case SERIAL_PARITY_EVEN:
			flags |= PARENB | PARODD;
			break;

		case SERIAL_PARITY_ODD:
			flags |= PARENB;
			break;

//		case SERIAL_PARITY_MARK:
//		case SERIAL_PARITY_SPACE:
		default:
			close(serial->fd);
			serial->fd = -1;
			return -1;
    	}
	if ((serial->fd = open(device, O_RDWR | O_NOCTTY)) < 0) {
		return -1;
	}
	printf("serial open %s fd: %d\n", device,serial->fd);

//printf("flags=%x; speed=%d;\n", flags, speed);
    	tcgetattr(serial->fd, &tconfig);
    	tconfig.c_cflag = flags;
    	tconfig.c_iflag = 0;
	tconfig.c_oflag = 0;
	tconfig.c_lflag = 0;
	tconfig.c_cc[VMIN] = 1;
	tconfig.c_cc[VTIME] = 0;
	cfsetispeed(&tconfig, speed);
	tcsetattr(serial->fd, TCSANOW, &tconfig);

	return 0;
}


int serial_close(serial_t *serial)
{
	close(serial->fd);
	serial->fd = -1;

	return 0;
}


void serial_release(serial_t *serial)
{
	if (serial->fd != -1) serial_close(serial);
	free(serial);
}

void serial_empty_data(serial_t *serial)
{
	lseek(serial->fd, 0, SEEK_END);
}

int serial_read(serial_t *serial, void *data, int length, long timeout)
{
	printf(" serial_has_read %d %d %d\n", serial->fd,length, timeout);
	struct timeval tv;
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(serial->fd, &fds);
	tv.tv_sec = 0;
	tv.tv_usec = timeout * 1000;
	if (timeout> 0) {
		if (select(serial->fd + 1, &fds, NULL, NULL, &tv) != 1) {
			return -1;
		}
	} else {
		if (select(serial->fd + 1, &fds, NULL, NULL, NULL) != 1) {
			return -1;
		}
	}

	int reb = read(serial->fd, data, length);
	int i;
	unsigned char* d = (unsigned char* ) data;
	printf(" ** serial_read rebut (%d/%d):",reb,length);
	for (i=0; i < reb; i++) printf(" %02x",d[i]);
	printf("\n");
	return reb;
}

int serial_write(serial_t *serial, void *data, int length)
{
	int i;
	int t;
	unsigned char* d = (unsigned char* ) data;
	printf(" ** serial_write (%d) enviant:", serial->fd);
	for (i=0; i < length; i++) printf(" %02x",d[i]);
	i = write(serial->fd, data, length);
	perror("HOLA");
	printf(" %s (%d)\n",i>0?"OK":"KO", i);
	return i;
}

/*
int serial_has_data(serial_t *serial, long millis)
{
	printf(" serial_has_data %d %d\n", serial->fd, millis);
	struct timeval tv;
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(serial->fd, &fds);
	tv.tv_sec = 0;
	tv.tv_usec = millis * 1000;
	if (millis> 0) {
		if (select(serial->fd + 1, &fds, NULL, NULL, &tv) == 1) {
			return 1;
		}
	} else {
		if (select(serial->fd + 1, &fds, NULL, NULL, NULL) == 1) {
			return 1;
		}
	}

	return 0;
}
*/

serial_config_t *serial_config_create(int speed, const char *options)
{
	struct serial_config cfg;
	serial_config_t *r;

    	if (!options) {
		return (serial_config_t *)malloc(sizeof(struct serial_config));
	}

	cfg.speed = speed;

	switch (options[0]) {
		case '7':
		case '8':
			cfg.word_size = options[0] - '0';
			break;

		default:
			return 0;
	}

	switch (options[1]) {
		case 'N':
		case 'P':
			break;

		default:
			return 0;
	}

	switch (options[2]) {
		case '1':
		case '2':
			cfg.stop_bits = options[2] - '0';
			break;

		default:
			return 0;
	}

	if (options[3] != 0) {
		return 0;
	}

printf("speed: %d\nword: %d\nstop: %d\n", cfg.speed, cfg.word_size, cfg.stop_bits);

	if (!(r = (serial_config_t *)malloc(sizeof(struct serial_config)))) {
		return 0;
	}

	memcpy(r, &cfg, sizeof(struct serial_config));

	return r;
}


void serial_config_release(serial_config_t *config)
{
	free(config);
}


char *serial_device(int port_number, int usb)
{
    char buf[30];

    sprintf(buf, usb?"/dev/ttyUSB%d":"/dev/ttyS%d", port_number - 1);

    return strdup(buf);
}
