#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <windows.h>
#include "serial.h"
#define CHARPTR(ptr)    ((char *)(ptr))
#define MIN(a, b) ((a) >= (b) ? (b) : (a))

#undef serial_config_create
#undef serial_has_data

struct serial {
	HANDLE h;
	serial_handler_t handler;
	OVERLAPPED read_overlapped;
	OVERLAPPED write_overlapped;
	int using_overlapped;
	char olbuf[1024];
	DWORD readed;
};


serial_t *serial_init()
{
	serial_t *s;

	if (!(s = (serial_t *)malloc(sizeof(struct serial)))) return 0;
	memset(s, 0, sizeof(struct serial));
	s->h = INVALID_HANDLE_VALUE;

	return s;
}


serial_handler_t serial_notification_handler(serial_t *serial, serial_handler_t handler)
{
	serial_handler_t old;

	old = serial->handler;
	serial->handler = handler;

	return old;
}


int serial_open(serial_t *serial, const char *device, serial_config_t *config)
{
	DCB dcb;
	DWORD dwError;
	BOOL fSuccess;
	COMMTIMEOUTS commTimeouts;
	HANDLE hCom;


	hCom = CreateFile(device,
		GENERIC_READ | GENERIC_WRITE,
		0,	 /* comm devices must be opened w/exclusive-access */
		NULL, /* no security attrs */
		OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_OVERLAPPED,	 /* not overlapped I/O */
		NULL  /* hTemplate must be NULL for comm devices */
	);

	if (hCom == INVALID_HANDLE_VALUE) {
		 logger("Error: unable to open port %s handle", device);
		 return -1;
	}

	logger("Seting Comm properties");
	if (!GetCommState(hCom, &dcb)) {
		 logger("Error: unable to retrieve port %s properties", device);
		 CloseHandle(hCom);
		 return -1;
	}

	dcb.BaudRate = config->speed;
	dcb.ByteSize = config->word_size;
	switch (config->parity) {
		case SERIAL_PARITY_NONE:
			dcb.Parity = NOPARITY;
			break;

		case SERIAL_PARITY_EVEN:
			dcb.Parity = EVENPARITY;
			break;

		case SERIAL_PARITY_ODD:
			dcb.Parity = ODDPARITY;
			break;

		case SERIAL_PARITY_MARK:
			dcb.Parity = MARKPARITY;
			break;

		case SERIAL_PARITY_SPACE:
			dcb.Parity = SPACEPARITY;
			break;

		default:
			logger("Error: Invalid parity (%d) at port %s.", config->parity, device);
			return -1;
	}

	switch (config->stop_bits) {
		case 1: dcb.StopBits = ONESTOPBIT; break;
		case 15: dcb.StopBits = ONE5STOPBITS; break;
		case 2: dcb.StopBits = TWOSTOPBITS; break;
		default:
			logger( "Error: Invalid stop bits (%f) at port %s. Valid values are 1 15 2", config->stop_bits, device);
			return -1;
	}

	if (!SetCommState(hCom, &dcb)) {
		 logger( "Error: unable to set port %s properties", device);
		 CloseHandle(hCom);
		 return -1;
	}

	logger( "Setting Comm Timeouts   ");
	if (!GetCommTimeouts(hCom, &commTimeouts)) {
		 logger("Error: unable to get port %s timeouts", device);
		 CloseHandle(hCom);
		 return -1;
	}

	commTimeouts.ReadIntervalTimeout=0;               //timeout between chars in ms
	commTimeouts.ReadTotalTimeoutMultiplier = 0;      //total timeout =
	commTimeouts.ReadTotalTimeoutConstant = 1500;  // ReadTotalTimeoutConstant + numchar*ReadTotalTimeoutMultiplier
	commTimeouts.WriteTotalTimeoutMultiplier = 0;
	commTimeouts.WriteTotalTimeoutConstant = 1500;

	if (!SetCommTimeouts(hCom, &commTimeouts)) {
		 logger("Error: unable to set port %s timeouts", device);
		 CloseHandle(hCom);
		 return -1;
	}

	logger("Serial: Connected to port %s, handle: %p", device, hCom);
	serial->h = hCom;

	SetCommMask(hCom, EV_RXCHAR);
	serial->using_overlapped = 1;
	serial->read_overlapped.Offset = 0;
	serial->read_overlapped.OffsetHigh = 0;
	serial->read_overlapped.hEvent = 0;
	serial->write_overlapped.Offset = 0;
	serial->write_overlapped.OffsetHigh = 0;
	serial->write_overlapped.hEvent = 0;

	return 0;
}


int serial_close(serial_t *serial)
{
	CloseHandle(serial->h);
	serial->h = INVALID_HANDLE_VALUE;

	return 0;
}


void serial_release(serial_t *serial)
{
	if (serial->h != INVALID_HANDLE_VALUE) serial_close(serial);
	free(serial);
}

int serial_write(serial_t *serial, void *data, int length)
{
	DWORD written;
	DWORD err;
	if (!WriteFile(serial->h, data, length, &written, &(serial->write_overlapped))) {
		err = GetLastError();
		if(err == ERROR_IO_PENDING) {
//			logger("Serial: Write Async\n");
			WaitForSingleObject(serial->h,INFINITE);
			err  = GetOverlappedResult(serial->h, &(serial->write_overlapped), &written,FALSE);
			FlushFileBuffers(serial->h);
//			logger("Serial: Wrote  %d of %d %x\n", written, length, err);
			return length;
		}
//		logger("Serial: Error: unable to write. error=%d", err);
		return -1;
	}

	return written;
}


int serial_read(serial_t *serial, void *data, int length, long timeout)
{
	DWORD readed;
	DWORD err;

	if (!ReadFile(serial->h, data, length, &readed, &(serial->read_overlapped))) {
		err = GetLastError();
		if(err == ERROR_IO_PENDING) {
//			logger("Serial: Read Async\n");
			if((err=WaitForSingleObject(serial->h, timeout?timeout:INFINITE)) != WAIT_OBJECT_0) {
//				logger("Serial: Read timeout! %x \n", err);
				CancelIo(serial->h);
			}
			GetOverlappedResult(serial->h, &(serial->read_overlapped), &readed,FALSE);
//			printf("Serial: Readed  %d of %d\n", readed, length);
			return readed;
		}
//		logger("Serial: Error: unable to read. error=%d", err);
		return -1;
	}

	return readed;
}


serial_config_t *serial_config_create(int speed, const char *options)
{
	struct serial_config cfg;
	serial_config_t *r;

    	if (!options) {
		return (serial_config_t *)malloc(sizeof(struct serial_config));
	}

	cfg.speed = speed;

	switch (options[0]) {
		case '7':
		case '8':
			cfg.word_size = options[0] - '0';
			break;

		default:
			return 0;
	}

	switch (options[1]) {
		case 'N':
		case 'P':
			break;

		default:
			return 0;
	}

	switch (options[2]) {
		case '1':
		case '2':
			cfg.stop_bits = options[2] - '0';
			break;

		default:
			return 0;
	}

	if (options[3] != 0) {
		return 0;
	}

printf("speed: %d\nword: %d\nstop: %d\n", cfg.speed, cfg.word_size, cfg.stop_bits);

	if (!(r = (serial_config_t *)malloc(sizeof(struct serial_config)))) {
		return 0;
	}

	memcpy(r, &cfg, sizeof(struct serial_config));

	return r;
}


void serial_config_release(serial_config_t *config)
{
	free(config);
}


char *serial_device(int port_number, int usb)
{
    char buf[30];

    sprintf(buf, "COM%d", port_number);

    return strdup(buf);
}
